package com.ebtlocator;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Comparator;
import java.util.Set;
import java.util.UUID;

import android.app.Activity;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.Bundle;
import android.os.CountDownTimer;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.EditText;
import android.widget.ListView;
import android.widget.Toast;
import android.widget.AdapterView.OnItemClickListener;

public class Beacon extends Activity {
	 
	   private Button listBtn;								//Show paired bluetooth devices list button
	   private Button findBtn;								//Show available bluetooth device in area bluetooth button
	   private Button beaconBtn;							//Turn on beacon mode button
	   private Button outBeaconBtn;							//Turn off beacon mode button
	   private Button calibrateBtn;							//Calibrate beacon button
	   private Button outCalibBtn;							//Reset calibration button
	   private Button selPosBtn;							//Confirm beacon position button
	   private EditText calibDist;							//Text editable field to insert distance of device and calibrate
	   private EditText beaconPos;							//Text editable field to insert the beacon position
	   private ListView myListView;							//List accessory of interface
	   private BluetoothAdapter myBluetoothAdapter;			//Bluetooth adapter to obtains self characteristic
	   private Set<BluetoothDevice> pairedDevices;			//Set that contains information of paired device adapters
	   private ArrayAdapter<String> BTArrayAdapter;			//Adapter between list view and other generated lists
	   private ArrayAdapter<BluetoothDevice> pairedList;	//Generated list of paired devices
	   private ArrayAdapter<DevLocalized> deviceList;		//Generated list of found devices
	   private ArrayAdapter<calibPoint> calibList;			//Generated list of calibration points
	   private BluetoothSocket mBluetoothSocket = null;		//Communication bluetooth socket
	   private int typeList = 0;							// 1: paired list, 2: Find list
	   private int indexMaster = 99;							//Master address into the paired devices list
	   private int rssiSelect = 0;							//Device rssi signal into unknown devices list
	   private int posX = 0;								//Position of any beacon in the X axis, initial origin x = 0
	   private boolean isCalibrated = false;
	   private UUID bluetoothUUID = UUID.fromString("00001111-0000-1000-18000-001011111111");	//Required key for socket connection						//
	   private String message = "blank";					//Initialization of interchanged message
	   
	   @Override
	   protected void onCreate(Bundle savedInstanceState) {
	      super.onCreate(savedInstanceState);
	      setContentView(R.layout.activity_beacon);

	      // take an instance of BluetoothAdapter - Bluetooth radio
	      myBluetoothAdapter = BluetoothAdapter.getDefaultAdapter();
	       
	          listBtn = (Button)findViewById(R.id.selMaster);
	          listBtn.setOnClickListener(new OnClickListener() {

	            @Override
	            public void onClick(View v) {
	                // TODO Auto-generated method stub
	            	typeList = 1;
	                list(v);
	            }
	          });

	          findBtn = (Button)findViewById(R.id.search);
	          findBtn.setOnClickListener(new OnClickListener() {

	            @Override
	            public void onClick(View v) {
	                // TODO Auto-generated method stub
	            	typeList = 2;
	                find();
	            }
	          });
	          
	          beaconPos = (EditText) findViewById(R.id.dist2Origin);
	          selPosBtn = (Button)findViewById(R.id.selecPos);
	          selPosBtn.setOnClickListener(new OnClickListener() {

	            @Override
	            public void onClick(View v) {
	            	if(beaconPos.getText().toString().length() > 0) {
	            		posX = Integer.parseInt(beaconPos.getText().toString());
	            		Toast.makeText(getApplicationContext(),"BALIZA EN X = " + posX,
                    			Toast.LENGTH_SHORT).show();
	            	}
	            }
	          });
	          
	          final CountDownTimer clientTimer = new CountDownTimer(11000, 2000) {
	      		 
	      	     public void onTick(long millis_to_end) {
	      	        find();
	      	        
	      	        if(message.length()>5) {
	      	        	
	      	        	ConnectThread conecta = new ConnectThread(pairedList.getItem(indexMaster));
	                  	if(mBluetoothSocket!=null) {
	                  		conecta.run();
	                  		conecta.cancel();
	                  	}
	                  	
	      	        }
	      	        	
	      	     }
	      	 
	      	     public void onFinish() {
	      	    	 this.start();
	      	     }
	      	  };
	          
	          beaconBtn = (Button)findViewById(R.id.startBeacon);
	          beaconBtn.setOnClickListener(new OnClickListener() {

	            @Override
	            public void onClick(View v) {
	            	typeList = 0;
	            	if(isCalibrated & indexMaster < 99)
	            		clientTimer.start();
	            	else { 	
	            		if(!isCalibrated)
	            			Toast.makeText(getApplicationContext(),"CALIBRACION SIN COMPLETAR",
	            					Toast.LENGTH_SHORT).show();
	            		if(indexMaster >= 99)
	            			Toast.makeText(getApplicationContext(),"SELECCIONAR MAESTRO",
	            					Toast.LENGTH_SHORT).show();
	            	}
	            }
	          });
	          
	          outBeaconBtn = (Button)findViewById(R.id.cancelBeacon);
	          outBeaconBtn.setOnClickListener(new OnClickListener() {

	            @Override
	            public void onClick(View v) {
	            	clientTimer.cancel();
	            }
	          });
	          
	          calibDist = (EditText) findViewById(R.id.distance);
	          calibrateBtn = (Button)findViewById(R.id.calib);
	          calibrateBtn.setOnClickListener(new OnClickListener() {

	            @Override
	            public void onClick(View v) {
	                // TODO Auto-generated method stub
	            	if(calibDist.getText().toString().length() > 0 && rssiSelect != 0) {
	                    calibList.add(new calibPoint(Integer.parseInt(calibDist.getText().toString()), rssiSelect));
	                    if(calibList.getCount()>1)
		                    calibList.sort(new Comparator<calibPoint>() {
								@Override
								public int compare(calibPoint o1, calibPoint o2) {
									return o1.compareTo(o2);
								}
		                    });
	                    for(int i = 0; i < calibList.getCount() ; ++i)
	                    	Toast.makeText(getApplicationContext(),"Calibrate : " + calibList.getItem(i).getDistance() + " " + calibList.getItem(i).getRSSI(),
	                    			Toast.LENGTH_SHORT).show();
	                    rssiSelect = 0;
	                    
	                }
	            	
	            	if(calibList.getCount()>2) {
	            		isCalibrated = true;
	            		Toast.makeText(getApplicationContext(),"Ya esta calibrado",
	                			Toast.LENGTH_SHORT).show();
	            	}	
	            	else
	            		isCalibrated = false;
	            }
	          });
	          
	          outCalibBtn = (Button)findViewById(R.id.resetCalib);
	          outCalibBtn.setOnClickListener(new OnClickListener() {

	            @Override
	            public void onClick(View v) {
	                // TODO Auto-generated method stub
	            	if(!calibList.isEmpty()) {
	                    calibList.clear();       
	                }
	            }
	          });

	          myListView = (ListView)findViewById(R.id.listView1);

//	           create the arrayAdapter that contains the BTDevices, and set it to the ListView
	          deviceList = new ArrayAdapter<DevLocalized>(this, android.R.layout.simple_list_item_1);
	          pairedList = new ArrayAdapter<BluetoothDevice>(this, android.R.layout.simple_list_item_1);
	          calibList = new ArrayAdapter<calibPoint>(this, android.R.layout.simple_list_item_1);
	          BTArrayAdapter = new ArrayAdapter<String>(this, android.R.layout.simple_list_item_1);
	          myListView.setAdapter(BTArrayAdapter);
	          
	          myListView.setOnItemClickListener(new OnItemClickListener() {
	  			
	  			@Override
	  			public void onItemClick(AdapterView<?> parent, View view, int position, long id) {
	  				// Mensaje emergente
	  				if(typeList == 1) {
	  					indexMaster = position;
	  					Toast.makeText(getApplicationContext(),"Dispositivo:"+"\n" + pairedList.getItem(position).getName() ,
	  							Toast.LENGTH_LONG).show();
	  				}
	  					
	  				if(typeList == 2) {
	  					rssiSelect = deviceList.getItem(position).getRSSI();
	  					Toast.makeText(getApplicationContext(),"RSSI = "+ deviceList.getItem(position).getRSSI(),
	  							Toast.LENGTH_LONG).show();
	  				}
	  					
	  			}
	  		});
	         
	      }
	   
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	                                      CLASS FUNCTIONS
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//PAIRED DEVICES LIST FUNCTION
	//OBTAINS BONDED DEVICES AND ADD EACH ONE (NAME AND ADDRESS) TO THE PAIREDLIST IN STRING FORMAT

	   public void list(View view){
	      pairedDevices = myBluetoothAdapter.getBondedDevices();

	      // put it's one to the adapter
	      BTArrayAdapter.clear();
	      pairedList.clear();
	      for(BluetoothDevice device : pairedDevices) {
	          BTArrayAdapter.add(device.getName()+ "\n" + device.getAddress());
	          pairedList.add(device);
	      }

	      Toast.makeText(getApplicationContext(),"Show Paired Devices",
	              Toast.LENGTH_SHORT).show();

	   }

	 //FOUND DEVICES LIST FUNCTION
	 //START DISCOVERY MODE OF THE BLUETOOTH ADAPTER AND INITIALIZE THE ARRAYS AND THE MESSAGE TO SEND
	 //ACTIVATE REGISTER-RECEIVER TO FOUND DEVICES IN AREA
	   
	   public void find() {
	       if (myBluetoothAdapter.isDiscovering()) {
	           // the button is pressed when it discovers, so cancel the discovery
	           myBluetoothAdapter.cancelDiscovery();
	       }
	       else {
	            BTArrayAdapter.clear();
	            deviceList.clear();
	            if(isCalibrated)
	            	message = "/";

	            myBluetoothAdapter.startDiscovery();

	            registerReceiver(bReceiver, new IntentFilter(BluetoothDevice.ACTION_FOUND));    
	        }    
	   }
	 
	 //BROADCAST RECEIVER (LISTENER) FUNCTION
	 //WAIT FOR INCOMMING DEVICE IN AREA (ACTION_FOUND)
	 //REGISTER THE DEVICE CHARACTERISTICS (EXTRA_DEVICE AND EXTRA_RSSI)
	 //STORAGE IN DEVICELIST IF IT NOT EXISTS IN THE LIST, IF IT EXISTS, ONLY CHANGE THE RSSI VALUE
	 //ADD INFORMATION AT THE MASSAGE TO SEND 
	   
	   final BroadcastReceiver bReceiver = new BroadcastReceiver() {
	        public void onReceive(Context context, Intent intent) {
	            String action = intent.getAction();

	            if (BluetoothDevice.ACTION_FOUND.equals(action)) {
	            	
	                 BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE);
			         int rssi = intent.getShortExtra(BluetoothDevice.EXTRA_RSSI,Short.MIN_VALUE);
			         boolean exist = false;
			         int pointer = 0;
			         int dist = 0;
			         
			         for(int i = 0; i <= deviceList.getCount()-1; ++i) {
			        	 if(deviceList.getItem(i).getBTDev().getAddress().equals(device.getAddress()) &
			        			 rssi!=deviceList.getItem(i).getRSSI() &
			        			 deviceList.getItem(i).getBTDev().getName().equals(device.getName())) {
			        		 exist = true;
			        		 pointer = i;
			        	 }
		             }
			         
			         if(exist) {
			        	 BTArrayAdapter.remove(BTArrayAdapter.getItem(pointer));
			        	 BTArrayAdapter.add(device.getName() + "\n" + device.getAddress() + "\n" + String.valueOf(rssi));
		                 BTArrayAdapter.notifyDataSetChanged();
		                 
		                 deviceList.remove(deviceList.getItem(pointer));
		                 deviceList.add(new DevLocalized(device, rssi));
			         }else if(!exist){
			        	 BTArrayAdapter.add(device.getName() + "\n" + device.getAddress() + "\n" + String.valueOf(rssi));
		                 BTArrayAdapter.notifyDataSetChanged();
			        	 deviceList.add(new DevLocalized(device, rssi));
			         }
			         
			         if(isCalibrated) {
				         dist = convert2Distance(rssi);
				         message = message + "%" + posX + "$" + device.getAddress() + "&" + dist + "/";
			         }
	            }
	        }
	    };
	   
	   public int convert2Distance(int rssi) {
		
		int distInt = 0;
		
		   for(int i = 0; i < calibList.getCount() ; ++i) {
			   if(rssi <= calibList.getItem(0).getRSSI()) {
				   calibPoint limitMax = new calibPoint(calibList.getItem(0).getDistance(),calibList.getItem(0).getRSSI());
				   calibPoint limitMin = new calibPoint(calibList.getItem(1).getDistance(),calibList.getItem(1).getRSSI());
				   
//				   //Interpolation between limitMax and limitMin from calibList
				   distInt = limitMin.getDistance() + (limitMax.getDistance()-limitMin.getDistance())/(limitMax.getRSSI()-limitMin.getRSSI())*(rssi - limitMin.getRSSI());
				   break;
			   }
			   else if(rssi > calibList.getItem(calibList.getCount()-1).getRSSI()) {
				   calibPoint limitMax = new calibPoint(calibList.getItem(calibList.getCount()-1).getDistance(),calibList.getItem(calibList.getCount()-1).getRSSI());
				   calibPoint limitMin = new calibPoint(calibList.getItem(calibList.getCount()-2).getDistance(),calibList.getItem(calibList.getCount()-2).getRSSI());
				   
//				   //Interpolation between limitMax and limitMin from calibList
				   distInt = limitMin.getDistance() + (limitMax.getDistance()-limitMin.getDistance())/(limitMax.getRSSI()-limitMin.getRSSI())*(rssi - limitMin.getRSSI());
				   break;
			   }
			   else if(rssi <= calibList.getItem(i).getRSSI()) {
				   calibPoint limitMax = new calibPoint(calibList.getItem(i-1).getDistance(),calibList.getItem(i-1).getRSSI());
				   calibPoint limitMin = new calibPoint(calibList.getItem(i).getDistance(),calibList.getItem(i).getRSSI());
				   
//				   //Interpolation between limitMax and limitMin from calibList
				   distInt = limitMin.getDistance() + (limitMax.getDistance()-limitMin.getDistance())/(limitMax.getRSSI()-limitMin.getRSSI())*(rssi - limitMin.getRSSI());
				   break;
			   }
		   }
		   return distInt;   
	   }

	 
	 //DESTROY FUNCTION
	 //DESTROY RECEIVER
	   
	   @Override
	   protected void onDestroy() {
	       // TODO Auto-generated method stub
	       super.onDestroy();
	       unregisterReceiver(bReceiver);
	   }

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	                                     COMMUNICATION CLASSES
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////
	   
	//CLIENT CLASS
	//CONSTRUCTOR :	CREATE A BLUETOOTH SOCKET
	//RUN			:	CONNECT WITH SERVER AND SEND MESSAGE
	   
	private class ConnectThread extends Thread {

	    public ConnectThread(BluetoothDevice device) {
	        try {
	            mBluetoothSocket = device.createRfcommSocketToServiceRecord(bluetoothUUID);
	        } catch (IOException e) {
	        	Toast.makeText(getApplicationContext(),"client NOT AVAILABLE" ,
	                    Toast.LENGTH_LONG).show();	        }
		    }
		    
	    public void run() {
	        try {
	            mBluetoothSocket.connect();
		        if (mBluetoothSocket != null) {
		             StreamThread mStreamThread = new StreamThread();
		             mStreamThread.start();
		                    
			         byte[] data = message.getBytes();        
			         mStreamThread.write(data);

		             Toast.makeText(getApplicationContext(), message,
			                 Toast.LENGTH_LONG).show();
		                 
		            }
		        } catch (IOException e) {
	            	Toast.makeText(getApplicationContext(),"Server NOT accept a client" ,
	                        Toast.LENGTH_LONG).show();
		        }
		    }
	    
		public void cancel() {
		    try {
		        mBluetoothSocket.close();
		    } catch (IOException e) {}
		}
		
	}
	 
	
	//STREAM CLASS
	//CONSTRUCTOR 	:	CREATE A COMMUNICATIONS STREAMS (BUFFERS) INPUT AND OUTPUT
	//READ			:	EXTRACT BYTES FROM IN BUFFER AND SAVE IN A STRING VARIABLE
	//WRITE	 		: 	WRITE BYTES IN THE OUT BUFFER

	private class StreamThread extends Thread {

	    private InputStream in;
	    private OutputStream out;

	    public StreamThread() {
	        try {
	            in = mBluetoothSocket.getInputStream();
	            out = mBluetoothSocket.getOutputStream();
	        } catch (IOException e) {}
	    }

	     public void write(byte[] data) throws IOException {
	    	try {
	    		out.write(data);
	           }catch (Exception e) {}
	     }
	}
	   
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//						                 DATA MANAGE CLASSES
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//FOUND DEVICES CLASS
	//CONSTRUCTOR 	:	INITIALIZE A DEVICE FROM BLUETOOTHDEVICE CLASS AND ITS RSSI SIGNAL
	//GETS			:	OBTAINS THE EXPLICIT OBJECT

	private class DevLocalized {

	   private BluetoothDevice btDev;
	   private int rssi;

	   public DevLocalized(BluetoothDevice btd, int signal) {
	      btDev = btd;
	      rssi = signal;
	   }

	   public BluetoothDevice getBTDev() {
		   return btDev;
	   }
	    	   
	   public int getRSSI() {
		   return rssi;
	   }

	}

	private class calibPoint{

		   private int distance;
		   private int rssi;

		   public calibPoint(int dist, int signal) {
		      distance = dist;
		      rssi = signal;
		   }

		public int compareTo(calibPoint o2) {
			// TODO Auto-generated method stub
			if (rssi < o2.rssi) {
	            return -1;
	        }
	        if (rssi > o2.rssi) {
	            return 1;
	        }
	        return 0;
		}

		public int getDistance() {
			   return distance;
		   }
		    	   
		   public int getRSSI() {
			   return rssi;
		   }
		}
	
}
